# ***********************************************************************
# * Copyright (c) infrae in 2002
# * For license details please see the license.txt file. 
# *
# * Project: Silva XML to Word Converter
# *
# * Creation Date: May 2002
# *
# * Authors: Stephan Richter
# ************************************************************************
"""
This module contains the code that handles a DOM that was generated from a
Silva export file.

$Id: SilvaDOMHandler.py,v 1.40 2003/03/27 10:56:21 zagy Exp $
"""

import string
import time
import os
import sys
import re
from types import ListType
from tempfile import mktemp
from win32com.client import Dispatch
from pywintypes import com_error
from util import reverseDict

import registry
from WordConstants import *
from word2silva.convert import p_LISTSTYLE
from StyleNameMapping import style_name_map
    
# XXX: replace by some configuration thingy
INDEX_TITLE = 'Index'
TOC_TITLE = 'Table of Contents'
PREFACE_FOLDER_IDS = ['leeswijzer' ,'preface']

p_SPACE = re.compile(r'\s+')

class SilvaDOMHandler:
    """Handles a Silva DOM tree and calls the right methods at the right time
       to create the document. This is a "technology-independent" component
       meaning that it will not depend on reportlab, latex or whatever.

       The above is not true. Yet :)
       
       If an unknown tag 'FOO' is encountered and no 'handle_FOO' method is
       available, it will load an approriate module registed in registry.py,
       and call the module's 'renderWord(self)'. 
       
       """

    _remove_from_p = ('\t', '\n')
    
    # defines the elements that can be inside a text element, such as 'p'
    _text_elements = (
        'link',
        'em', 
        'strong', 
        'index', 
        'underline',
        'super', 
        'sub',
        'br',
    ) + tuple(registry.getTagsOfType('text'))
   
    # elements, contained in a doc
    _container_elements = (
        'list', 
        'nlist', 
        'p', 
        'heading', 
        'table', 
        'dlist',
        'pre',
        ) + tuple(registry.getTagsOfType('container'))

    _toplevel_elements = (
        'silva_publication', 
        'silva_folder',
        'silva_document'
    ) + tuple(registry.getTagsOfType('toplevel'))
    
    _align_mapping = {
        'L': 0,
        'R': 2,
        'C': 1,
        'J': 5
    }


    _list_xml = {
        'bullet': 'List Bullet',
        'disc': 'List Bullet',
        'circle': 'List Circle',
        'square': 'List Square',
        'dash': 'List Dash',
        'none': 'List',
        '1': 'List Number',
        'a': 'List Letter',
        'A': 'List Letter Upper', 
        'i': 'List Roman',
        'I': 'List Roman Upper',
    }

    _list_word = reverseDict(_list_xml)


    def __init__(self, dom, titledoc='', template='', visibility = 2):
        """Initialize the handler."""

        self.titledoc = titledoc
        self.template = template
        self.visibility = visibility
        self.docData = None
        
        self.output = ''
        self.level = 0
        self.currNode = None
        self.list_level = 0
        self.list_type = 10

        self.useParagraphStyle = 'Body Text'
        self.setStyle = 1
        self.do_newline = 1

        self.isFirstContentPage = False

        self.app = Dispatch("Word.Application")
        self.app.Visible = (visibility >= 2)
        
        self.docName = mktemp('.doc')
        
        if titledoc:
            self.doc = self.app.Documents.Open(os.path.abspath(titledoc), 
                ReadOnly=1)
            if template:                
                self.doc.AttachedTemplate = template
        else:
            self.doc = self.app.Documents.Add(Template=template)
        self.doc.SaveAs(self.docName)
    
    
        silva_node = dom.firstChild
        childs = filter(lambda x: x.nodeType == x.ELEMENT_NODE, 
            silva_node.childNodes)
        if len(childs) == 1:
            if childs[0].nodeName == 'silva_document':
               self.currNode = silva_node
            else:
                self.currNode = childs[0]
        else:
            self.currNode = silva_node
        self._cacheListIndentation()
        self._setupListStyles()
        self.__list_isNlist = []
        self.getStyle('Hyperlink').Font.Underline = 0

    def _reboot_word(self):
        self.doc.Close(SaveChanges=wdSaveChanges)
        self.app.Quit(SaveChanges=wdDoNotSaveChanges)
        self.app = None

        self.app = Dispatch("Word.Application")
        self.doc = self.app.Documents.Open(self.docName)
        self.app.Visible = (self.visibility >= 2)

        self.app.Selection.Start = self.app.Selection.End = \
            self.doc.Range().End

    def _cacheListIndentation(self):
        """caches the LeftIndent attribute of all list styles"""
        
        cache = {}

        level = 0
        while 1:
            level += 1
            name = 'List'
            if level > 1:
                name += ' %i' % (level, )
            try:                
                style = self.getStyle(name, autoAdd=False)
            except KeyError:                 
                break
            cache[level] = style.ParagraphFormat.LeftIndent
        self.__list_indentation = cache            
        
        
    def _setupListStyles(self):
        """fixes the indentation of all list styles"""

        max_level = max(self.__list_indentation.keys())
        for base_name in self._list_word:
            for level in range(1, max_level+1):
                if level == 1: 
                    name = base_name
                else: 
                    name = '%s %s' % (base_name, level)
                self.getStyle(name).ParagraphFormat.LeftIndent = \
                    self.__list_indentation[level]
           
    def __del__(self):
        if getattr(self, 'dontClose', False):
            return
        if getattr(self, 'doc', None) is not None:
            try:
                self.doc.Close(SaveChanges=True)
            except com_error:
                pass
        try:
            os.unlink(self.docname)
        except AttributeError:
            pass
        except OSError:
            pass
            
    def getNodeAttribute(self, node, name):
        """Return value of node attribute."""
        return node.attributes.get(name).value
    
    def addComment(self, Range=None, **kw):
        """Add a word comment to the document.
        
            Range: a Range object to which the commend should be added. 
                If None, app.Selection is used.
        """
        text = ';'.join(
            map(lambda (x, y): '%s="%s"' % (x, y), 
            map(lambda (x, y): (x, y.replace('"', '\"')),
            map(lambda (x, y): (str(x), y), 
                kw.items()))))
        comment = 'slv:%s' % (text, )
        if Range is None:
            Range = self.app.Selection.Range
        Range.Comments.Add(Range, comment)
    
    def _getText(self, node):
        """Retrieves all the contained text from a node."""        
        text = u""
        for child in node.childNodes:
            if child.nodeType == child.TEXT_NODE:
                text += child.data
        text = text.strip()
        words = text.split()
        if words:
            text = ' '.join(words)
        return text


    def _getTextOfTag(self, tagName, node=None):
        """Retrieves the text of a certain element having tagName"""
        if node is None:
            node = self.currNode
        nodes = node.getElementsByTagName(tagName)
        if len(nodes) > 0:
            return self._getText(nodes[0])
        return ""


    def _walkSubNodes(self, tagNames=()):
        """Walks through all the specified sub nodes of the current node and
           handles them. """
        currNode = self.currNode
        result = []

        for child in currNode.childNodes:
            name = getattr(child, 'tagName', None)
            if name in tagNames:
                self.currNode = child
                sub_res = self._execNodeHandler()
                if type(sub_res) == ListType:
                    result += sub_res
                else:
                    result.append(sub_res)
                
        self.currNode = currNode
        return result
    
    def _execNodeHandler(self):
        """Execute the handler of the current node."""
        tagName = getattr(self.currNode, 'tagName', None)
        if tagName is not None:
            self.level += 1
            handler = getattr(self, 'handle_'+tagName, None)
            if handler:
                result = handler()
            else:
                # load plugin for unknown tag
                try:
                    result = registry.getPlugin(tagName).renderWord(self)    
                except KeyError:
                    result = []
            self.level -= 1
            return result


    def _has_node_valid_subs(self, node):
        return filter(lambda node, s=self: node.nodeType != 3 and \
                      node.tagName in self._container_elements,
                      node.childNodes)
        

    def _insertTOC(self):
        """Adds TOC to to story"""
        # This might be needed, but not right now as the colophon fills
        # already the page
        # self.app.Selection.InsertBreak()
        self.app.Selection.Style=self.getStyle('TOC Heading')
        self.app.Selection.TypeText(TOC_TITLE+'\n')

        self.app.Selection.Style=self.getStyle(self.useParagraphStyle)
        self.app.Selection.TypeText(' ')
        text_len = self.doc.Range().End
        self.toc = self.doc.TablesOfContents.Add(
            self.doc.Range(text_len-1, text_len))

        # Clean-up
        end = self.doc.Range().End
        self.app.Selection.Start = end
        self.app.Selection.End = end


    def _insertIndex(self):
        """Add index to story"""
        self.app.Selection.InsertBreak()
        self.app.Selection.Style=self.getStyle('Heading 1')
        self.app.Selection.TypeText(INDEX_TITLE+'\n')

        self.app.Selection.Style=self.getStyle(self.useParagraphStyle)
        self.app.Selection.TypeText(' ')
        text_len = self.doc.Range().End
        index = self.doc.Indexes.Add(
            self.doc.Range(text_len-1, text_len))
        index.NumberOfColumns=2


    def _fixSections(self):
        """Disconnect section 2 from section 1."""
        if self.doc.Sections.Count < 2:
            print "Not disconnecting sections since there's only one"
            return
        
        for x in range(1, 4):
            self.doc.Sections(2).Headers(x).LinkToPrevious = 0
            footer = self.doc.Sections(2).Footers(x)
            footer.LinkToPrevious = 0
            footer.PageNumbers.RestartNumberingAtSection = 1

        for x in range(1, 4):
            self.doc.Sections(1).Headers(x).Range.Delete()
            self.doc.Sections(1).Footers(x).Range.Delete()
    
    def getListStyleName(self, base=None, level=None):
        """returns the list style based on base for given level
        
            if base is None self._list_xml[self.list_type] is used
            if level is None self.list_level is used
            
            raises ValueError if level < 1
        """
        if base is None: base = self._list_xml[self.list_type]
        if level is None: level = self.list_level
        if level < 1:
            raise ValueError, "%i is not a valid list level" % (level, )
        if level == 1:
            return base
        return '%s %i' % (base, level)


    def getStyle(self, name, autoAdd=True):
        
        try: word_style = style_name_map[name]
        except KeyError: word_style = name
            
        try: 
            return self.doc.Styles(word_style)
        except com_error: 
            if autoAdd:
                # the requested style is not in the document, so we add it
                style = self.doc.Styles.Add(word_style, wdStyleTypeParagraph)
                if name.startswith('List'): base = wdStyleList
                else: base = wdStyleBodyText
                style.BaseStyle = base
                return style
            else:
                raise KeyError, 'Style %s is not available' % (name, )



    def __getStyle(self, name, autoAdd=True):
        requested_name = name
        
        if self.list_level > 0 and not name.startswith('List'):
            list_style_name = self.getListStyleName('List continue')
            name = '%s %s' % (list_style_name, name)
            style = None
            try: style = self._getStyle(name, autoAdd=False)
            except KeyError: pass
            if style is None:
                list_style = self._getStyle(list_style_name)
                style = self.doc.Styles.Add(name, wdStyleTypeParagraph)
                style.BaseStyle = self._getStyle(requested_name)
                style.ParagraphFormat.LeftIndent = \
                    list_style.ParagraphFormat.LeftIndent
        else:
            style = self._getStyle(name, autoAdd)

        return style

    def _clear_table(self, table):
        for x in range(1, 7):
            table.Borders(x).Visible = 0


    def _make_stats(self, start, end):
        delta_time = end-start
        mins = int(delta_time)/60
        secs = delta_time%60

        print "Required time: %.2i:%.2d" %(mins, secs)
        print "Pages Generated: %i" %self.doc.Range().ComputeStatistics(2)


    def _setProperties(self):

        prop = self.doc.BuiltinDocumentProperties
        prop(wdPropertyTitle).value = self.title
        prop(wdPropertyAuthor).value = u'Silva2Word'
        prop(wdPropertyCompany).value = u''
        prop(wdPropertyComments).value = u'This file was generated '\
            'by Silva2Word.\nslv:id="%s";obj="%s"' % (self.rootId, 
                self.rootObject)


    def _updateFields(self):
   
        self.doc.Fields.Update()

    def run(self):
        """Process the dom tree."""
        start = time.time()
        try:            
            docid = self.currNode.attributes.get('id').value
        except AttributeError:
            docid = None
        if not docid:
            docid = 'silva_root'
        self.rootId = docid            
        self.title = title = self._getTextOfTag('title')
        docobj = self.currNode.nodeName
        # you cannot import a root anyway, so throw it away
        if docobj == 'silva_root': 
            docobj = 'silva_publication'
        self.rootObject = docobj
        self._setProperties()

        self.isFirstContentPage = True
        self.app.Selection.Start = self.doc.Range().End
        self.app.Selection.End = self.doc.Range().End
        self._walkSubNodes(self._toplevel_elements)
        
        self._insertIndex()        
        self._fixSections()
        self._updateFields() 
        end = time.time()
        self.app.Visible = (self.visibility >= 1)
        self._make_stats(start, end)
        
    def getDOC(self):
        """return document"""

        if self.docData is None:
            if self.doc is None:
                return None
            self.doc.Save()
            self.doc.Close()
            self.doc = None
    
            f = open(str(self.docName), "rb")
            self.docData = f.read()
            f.close()
            
        return self.docData
        
    def handle_silva_publication(self):
        """Handle silva_publication element."""
        self.handle_silva_folder()

    def handle_silva_root(self):
        """Handle silva_publication element."""
        self.handle_silva_folder()


    def _handle_preface_silva_folder(self):
        """Handle silva_folder element with id in PREFACE_FOLDER_IDS"""

        # Place cursor to the right place to insert the preface
        self.app.Selection.Start = self.doc.Range().Start
        self.app.Selection.End = self.doc.Range().End
        find = self.app.Selection.Find
        find.ClearFormatting()
        find.Text = u''
        find.MatchCase = 0
        find.MatchWholeWord = 0
        find.MatchWildcards = 0
        find.MatchSoundsLike = 0
        find.MatchAllWordForms = 0
        find.Style = 'TOC Heading'
        find.Execute()
        
        start = self.app.Selection.Start 
        self.app.Selection.End = start
        self.app.Selection.InsertBreak()
        self.app.Selection.End = start

        # Now handle the content
        id = self.currNode.attributes.get('id').value
        tagname = self.currNode.nodeName
        self.addComment(id=id, obj=tagname)
        title = self._getTextOfTag('title')
        self.app.Selection.Style=self.getStyle('Preface Heading')
        self.app.Selection.TypeText(title+'\n')

        self._walkSubNodes(self._toplevel_elements)

        self.app.Selection.Style = self.getStyle('Normal')
        # Place the cursor on the right position (end of document)
        end = self.doc.Range().End
        self.app.Selection.Start = end
        self.app.Selection.End = end
        
        

    def handle_silva_folder(self):
        """Handle silva_folder element."""
        # There is this error message in word `The formatting in this 
        # document is too compelex. Pleas full save the document now.'
        # Folders appear regularly, but not too often, so save here.
        self._reboot_word()
        
        # this should be one of silva_folder, silva_root, silva_publication
        tagname = self.currNode.nodeName 
        
        id = self.currNode.attributes.get('id').value
        if id.lower() in PREFACE_FOLDER_IDS:
            return self._handle_preface_silva_folder()
        
        self.handle_title()
        self.addComment(id=id, obj=tagname)
        self._walkSubNodes(self._toplevel_elements)
        

    def handle_silva_document(self):
        """Handle silva_document element."""
        id = self.currNode.attributes.get('id').value
        tagname = self.currNode.nodeName

        if id != 'index':
            self.handle_title()
            self.addComment(id=id, obj=tagname)
        
        self.currNode = self.currNode.getElementsByTagName('doc')[0]
        self._walkSubNodes(self._container_elements)


    def handle_field(self):
        """Handle field element (inside row --> table)."""
        self.curr_col += 1
        start = self.curr_table.Cell(self.curr_row, self.curr_col).Range.Start
        self.app.Selection.Start = start
        self.app.Selection.End = start
        if self._has_node_valid_subs(self.currNode):
            setStyle = self.setStyle
            paragraphStyle = self.useParagraphStyle
            self.useParagraphStyle = 'Table Body Text'
            self.setStyle = 0
            self._walkSubNodes(self._container_elements)
            self.setStyle = setStyle
            self.useParagraphStyle = paragraphStyle
        else:
            self.renderParagraph(0, 0)

        if self.curr_table.Cell(
            self.curr_row, self.curr_col).Range.Paragraphs.Count == 1:
            self.curr_table.Cell(self.curr_row, self.curr_col
                                 ).Range.Paragraphs(1).Alignment = \
                self._align_mapping[self.col_aligns[self.curr_col-1]]
            

    def handle_column_heading(self):
        """Handle column_heading element (inside row --> table)."""
        self.handle_field()
        self.curr_table.Cell(self.curr_row, self.curr_col).Range.Style = \
                                            self.getStyle('Column Heading')
        

    def handle_row_heading(self):
        """Handle row element (inside table)."""
        self.curr_row += 1
        if self.curr_row != 1:
            self.curr_table.Rows.Add()
        self.curr_table.Rows(self.curr_row).Cells.Merge()
        self.curr_table.Cell(self.curr_row, 1).Range.Style = \
               self.getStyle('Row Heading')
        start = self.curr_table.Cell(self.curr_row, 1).Range.Start
        self.app.Selection.Start = start
        self.app.Selection.End = start
        self.renderParagraph(0, 0, 1)
        self.after_row_heading = 1


    def handle_row(self):
        """Handle row element (inside table)."""
        self.curr_row += 1
        if self.curr_row != 1:
            self.curr_table.Rows.Add()
        if self.after_row_heading:
            self.curr_table.Rows(self.curr_row).Cells.Split(1, self.table_cols)
            self.after_row_heading = 0
        self.curr_table.Rows(self.curr_row).Range.Style = \
                                          self.getStyle('Table Body Text')
        self.curr_col = 0
        self._walkSubNodes(('field', 'column_heading'))


    def handle_table(self):
        """Handle table element."""        
        # Find the table type
        table_type = self.currNode.attributes.get('type', None)
        if table_type is not None:
            table_type = table_type.value
        else:
            table_type = 'plain'

        # Special customer request: Add an Empty line before grid tables.
        if table_type == 'grid':
            self.app.Selection.Style = self.getStyle('Body Text')
            self.app.Selection.TypeText('\n')

        cols = int(self.currNode.attributes.get('columns').value)
        self.table_cols = cols
        self.after_row_heading = 0

        self.app.Selection.TypeText(' ')
        text_len = self.doc.Range().End
        self.curr_table = self.app.Selection.Tables.Add(
            self.doc.Range(text_len-2, text_len), 1, cols)

        # Handle column widths
        table_width = self.curr_table.Columns.Width*cols
        col_info = self.currNode.attributes.get('column_info', None)
        # This has to be done in case we do not have colsizes.
        if col_info is not None:
            colsizes = map(lambda x: x.split(':')[1],
                           col_info.value.split(' '))
            self.col_aligns = map(lambda x: x.split(':')[0],
                                  col_info.value.split(' '))
        else:
            colsizes = ['1']*cols
            self.col_aligns = ['L']*cols
        widths = map(float, colsizes)
        unitWidth = (table_width)/reduce(lambda x, y: x+y, widths)
        widths = map(lambda x: x*unitWidth, widths) 

        for i in range(0, len(widths)):
            self.curr_table.Columns[i].Width = widths[i]

        # Handle all rows
        self.curr_row = 0
        result = self._walkSubNodes(('row', 'row_heading'))         

        # Apply style
        self.curr_table.LeftPadding = 1.5
        self.curr_table.RightPadding = 1.5
        self.curr_table.TopPadding = 0.5
        self.curr_table.LeftPadding = 0.5
            
        self._clear_table(self.curr_table)
        if table_type == 'plain':
            pass
        elif table_type == 'grid':
            for b in range(6):
                self.curr_table.Borders(b+1).Visible = 1
            self.curr_table.Rows(1).Range.Bold = 1
            self.curr_table.Rows(1).HeadingFormat = -1
            
        elif table_type == 'datagrid':
            for b in range(6):
                self.curr_table.Borders(b+1).Visible = 1
            self.curr_table.Rows(1).Borders(1).Visible = 0
            self.curr_table.Rows(1).Borders(2).Visible = 0
            self.curr_table.Rows(1).Borders(4).Visible = 0
            self.curr_table.Rows(1).Borders(5).Visible = 0
            self.curr_table.Rows(1).Range.Bold = 1

        elif table_type == 'listing':
            self.curr_table.Rows(1).Borders(3).Visible = 1
            self.curr_table.Rows(1).Range.Bold = 1

        #Tag the first table cell with the style attributes.
        self.addComment(Range=self.curr_table.Cell(1,1).Range, type=table_type)

        # Clean up. Bring us back to the end of the document
        self.app.Selection.Start = self.curr_table.Range.End
        self.app.Selection.End = self.curr_table.Range.End

        self.app.Selection.Style=self.getStyle('One Pixel')
        self.app.Selection.TypeText('\n')

        # I think we should not do this
        # self.app.Selection.Style = self.getStyle(self.useParagraphStyle)
        # self.app.Selection.TypeText("\n")
        

    def handle_nlist(self):
        """Handle nlist element."""
        self.__list_isNlist.append(True)
        self.__handle_list()
        self.__list_isNlist.pop()


    def handle_list(self):
        """Handle list element."""
        self.__list_isNlist.append(False)
        self.__handle_list()
        self.__list_isNlist.pop()
        
       
    def __handle_list(self):
        """worker for nlist and list"""
        self.list_level += 1
        prev_list_type = self.list_type
        try:
            list_type = self.currNode.attributes.get('type').value
        except:
            list_type = 'none'
        self.list_type = list_type
        start = self.app.Selection.End
        self._walkSubNodes(('li', ))
        end = self.app.Selection.End
        r = self.doc.Range(start, end)
        wp = r.Paragraphs[0]
        lf = wp.Range.ListFormat
        try:
            lf.ApplyListTemplate(lf.ListTemplate, ContinuePreviousList=False)
        except com_error:
            pass
        self.app.Selection.Start = self.app.Selection.End = \
            self.doc.Range().End
        self.app.Selection.Style = self.getStyle("List Spacing")
        self.app.Selection.TypeText('\n')
        self.app.Selection.Start = self.app.Selection.End = \
            self.doc.Range().End
        self.app.Selection.Style = self.getStyle("One Pixel")
        self.list_type = prev_list_type
        self.list_level -= 1

    def handle_li(self):
        """Handle li element (inside list)."""
        old_paragraph_style = self.useParagraphStyle
        if self.__list_isNlist[-1]:
            li_node = self.currNode
            node = li_node.firstChild
            first_element = None
            while node is not None:
                self.currNode = curr_node = node
                node = node.nextSibling
                if curr_node.nodeType == curr_node.ELEMENT_NODE:
                    if first_element is None:
                        first_element = curr_node
                else:                        
                    continue
                if first_element is curr_node:
                    if curr_node.nodeName == 'p':
                        self.useParagraphStyle = self.getListStyleName()
                        self.renderParagraph(setStyle=True)
                        continue
                    else:
                        self.app.Selection.Style = self.getStyle(
                            self.getListStyleName())
                        self.app.Selection.TypeText('\n')
                start_index = self.app.Selection.End
                self._execNodeHandler()
                end_index = self.app.Selection.End
                r = self.doc.Range(start_index, end_index)
                indent = self._getLiIndent(r.ParagraphFormat.LeftIndent)
                try:
                    r.ParagraphFormat.LeftIndent = indent
                except AttributeError:
                    pass
            self.currNode = li_node                    
        else:
            self.useParagraphStyle = self.getListStyleName(
                self._list_xml[self.list_type])
            self.renderParagraph(setStyle=True)
        self.useParagraphStyle = old_paragraph_style
               
    def _getLiIndent(self, current_indent):
        
        this_indent = self.__list_indentation.get(self.list_level)
        next_indent = self.__list_indentation.get(self.list_level+1, 
            sys.maxint)
        if this_indent is None:
            raise ValueError, "Lists are nested too deep (curr=%i, max=%i)" % (
                self.list_level, max(self.__list_indentation.keys()))
        new_indent = current_indent + this_indent
        if new_indent > next_indent: 
            new_indent -= 1
        return new_indent            

    def handle_p(self):
        if self.setStyle:
            self.useParagraphStyle = 'Body Text'
        self.renderParagraph(do_newline=1, setStyle=True)
        
    def renderParagraph(self, do_newline=None, setStyle=None, do_actions=1):
        """Handle p element."""

        if do_newline is None:
            do_newline = self.do_newline

        if setStyle is None:
            setStyle = self.setStyle

        p_start = self.app.Selection.End
        
        p = self.currNode
        actions = []

        for child in self.currNode.childNodes:
            if child.nodeType == child.TEXT_NODE:
                text = child.data
                text = p_SPACE.sub(' ', text)
                if text:
                    self.app.Selection.TypeText(text)
            tagName = getattr(child, 'tagName', None)
            if tagName in self._text_elements:
                self.currNode = child
                start = self.app.Selection.End
                actions += self.renderParagraph(0, 0, 0)
                end = self.app.Selection.End
                if tagName == 'em':
                    actions.append((start, end, 'Italic'))
                elif tagName == 'strong':
                    actions.append((start, end, 'Bold'))
                elif tagName == 'underline':
                    actions.append((start, end, 'Underline'))
                elif tagName == 'super':
                    actions.append((start, end, 'Superscript'))
                elif tagName == 'sub':
                    actions.append((start, end, 'Subscript'))
                elif tagName == 'link':
                    url = child.attributes['url'].value
                    if url == '':
                        url = '#'
                    r = self.doc.Range(start, end)
                    try:
                        self.doc.Hyperlinks.Add(r, url)
                    except com_error:
                        # XXX: catch only 'Bad parameter'
                        # Bad parameter is raised if the url parameter is 
                        # not an url.
                        pass
                elif tagName == 'index':
                    name = '"%s"' % (child.attributes['name'].value, )
                    r = self.doc.Range(end, end)
                    try:
                        self.app.Selection.Fields.Add(r, wdFieldIndexEntry, 
                            name)
                    except com_error:
                        # XXX: sometimes word returns 'Command not available'
                        # and i haven't got a clue why.
                        pass
                elif tagName == 'br':
                    self.app.Selection.TypeText('\x0b')

        if do_newline:
            try: newlines = int(do_newline)
            except TypeError: newlines = 1
            self.app.Selection.TypeText(newlines*"\n")

        self.currNode = p

        p_end = self.doc.Range().End
        
        if setStyle:
            self.doc.Range(p_start, p_end).Style = \
                self.getStyle(self.useParagraphStyle)
            
        if do_actions:
            for action in actions:
                setattr(self.doc.Range(action[0], action[1]).Font, action[2], 1)
        else:
            return actions
      

    def handle_title(self):
        """Insert a title to the story."""
        title = self._getTextOfTag('title')
        if title:
            output = ''
            if self.level == 1: 
                if self.isFirstContentPage:
                    self.isFirstContentPage = False
                else:
                    self.app.Selection.InsertBreak()

            self.app.Selection.Style=self.getStyle('Heading %i' %self.level)
            self.app.Selection.TypeText(title+'\n')


    def handle_heading(self):
        """Handle heading element."""
        type = self.currNode.attributes.get('type')
        if type is None:
            type = 'Normal'
        else:
            type = type.value.capitalize()

        self.app.Selection.Style = self.getStyle('%s Heading' %type)
        self.renderParagraph(1, 0, 1)


    def handle_dlist(self):
    
        # XXX: normal or compact, not handled yet
        style = self.currNode.attributes.get('type').value 
        
        node = self.currNode.firstChild
        while node is not None:
            self.currNode = node
            if node.nodeName == 'dt':
                self.useParagraphStyle = 'Definition Term'
                self.renderParagraph(1, 1, 1)
            elif node.nodeName == 'dd':
                self.useParagraphStyle = 'Definition Description'
                self.renderParagraph(1, 1, 1)

            node = node.nextSibling
            
        self.currNode = self.currNode.parentNode

    def handle_pre(self):
   
        node = self.currNode
        
        text = u""
        for child in node.childNodes:
            if child.nodeType == child.TEXT_NODE:
                text += child.data
        
        start = self.app.Selection.End
        self.app.Selection.TypeText(text+'\n')
        end = self.app.Selection.End

        r = self.doc.Range(start, end)
        style = self.getStyle('Plain Text')
        r.Style = style
        
        
