# Copyright (c) 2002 Infrae. All rights reserved.
# See also LICENSE.txt
# $Id: render.py,v 1.2 2002/12/24 14:07:47 zagy Exp $



"""Rendering library for Silva.

This contains various utilty classes for rendering Silva documents.

Plugin modules for rendering specific Silva tags can use these classes
for rendering, but they don't need to. The classes are designed so that
they can support multiple rendering formats. At the moment however,
they only support MS Word rendering.
"""


class Renderable:
    """An object that can be rendered on to a document."""
    
    postProcessWord = None
    
    def setPostProcessWord(self, f):
        """Set a post processing function for Word COM objects.

        f will be called with two arguments - f(handler, obj):
           handler - the SilvaDOMHandler used for rendering
           obj - the Word COM object.

        This method allows you to customize output of Renderables, without
        having to subclass them.
        """
        assert callable(f)
        self.postProcessWord = f
    
    def renderWord(self, handler):
        """Render to Word, given a SilvaDOMHandler object.

        Override this if you don't want the default policy.
        """
        obj = self.doRenderWord(handler)
        if self.postProcessWord:
            self.postProcessWord(handler, obj)
    
    def doRenderWord(self, handler):
        """Should return a Word COM object that was created."""
        raise NotImplementedError, "implement in subclasses"

    def cleanupWord(self, handler):
        """Move to end of the document."""
        end = handler.doc.Range().End
        sel = handler.app.Selection
        sel.Start = end
        sel.End = end
        sel.Style = handler.getStyle("One Pixel")


class Table(Renderable):
    """A table."""

    def __init__(self, subtable=0):
        self.rows = []
        self.widths = None
        self.subtable = subtable
    
    def setWidths(self, *widths):
        """Set the widths of the table columns.

        Values should be floating point values that add up to 1.
        """
        self.widths = widths
    
    def addRow(self, *cells):
        """Add a row of cells to the table.

        Cells should be Cell instances.
        """
        self.rows.append(cells)

    def doRenderWord(self, handler):
        app, doc = handler.app, handler.doc
        if self.subtable:
            app.Selection.Style = handler.getStyle(handler.useParagraphStyle)
            app.Selection.TypeText(' ')
            text_len = app.Selection.End
        else:
            app.Selection.TypeText(' ')
            text_len = doc.Range().End
        table = app.Selection.Tables.Add(
            doc.Range(text_len-1, text_len), len(self.rows), max(map(len, self.rows)))

        handler._clear_table(table)
        
        # set table widths
        table_width = table.Columns.Width*2
        if self.widths:
            for i in range(len(self.widths)):
                table.Columns(i+1).Width = table_width * self.widths[i]

        # create cell contents
        i = 0
        for row in self.rows:
            j = 0
            for cell in row:
                start = table.Cell(i+1, j+1).Range.Start
                handler.app.Selection.Start = start
                handler.app.Selection.End = start
                for r in cell.renderables:
                    r.renderWord(handler)
                if cell.style:
                    table.Cell(i+1,j+1).Range.Style = handler.getStyle(cell.style)
                j += 1
            i += 1
        return table


class Cell:
    """A cell in a table.

    Accepts a list of renderables and an optional style for the cell.
    """
    
    def __init__(self, renderables, style=None):
        self.renderables = renderables
        self.style = style


class Text:
    """Render text."""

    def __init__(self, text):
        self.text = text

    def renderWord(self, handler):
        handler.app.Selection.TypeText(self.text)


class FieldText:
    """Store text taken from a field."""

    def __init__(self, field):
        self.field = field
    
    def renderWord(self, handler):
        handler.app.Selection.TypeText(handler._getTextOfTag(self.field))        


class Record(Renderable):
    """A record, composed of record fields."""
    
    def __init__(self):
        self.contents = []
        self.construct()
        
    def add(self, renderable):
        """Add a new item to be rendered."""
        self.contents.append(renderable)
        return renderable
    
    def renderWord(self, handler):
        handler.app.Selection.InsertBreak(Type=0)
        for r in self.contents:
            r.renderWord(handler)
            self.cleanupWord(handler)
        del self.contents

    def construct(self):
        """Actual construction - factory method."""
        raise NotImplementedError, "implement in subclasses"


__all__ = "Renderable Table Cell Text FieldText Record".split()
