# ************************************************************************
# * Copyright (c) infrae in 2002
# * For license details please see the license.txt file. 
# *
# * Project: DOC to Silva XML Converter
# *
# * Creation Date: June 2002
# *
# * Authors: Stephan Richter, Christian Zagrodnick
# ************************************************************************
"""

$Id: P.py,v 1.18 2003/03/17 11:44:01 zagy Exp $
"""

import re
from copy import copy

from util import replaceMap, stripControlChars
from silva2word.WordConstants import *
from word2silva.SilvaObject import SilvaObject

p_INDEX = re.compile(r' "([^"]+)" ')
p_SoftBreak = re.compile('\x0b+')


class Tag:
    
    def __init__(self, name, attributes, on, position):
        self.name = name
        self.attributes = attributes
        self.on = on
        self.position = position

    def tag(self):
        space = ''
        empty = ''
        if self.on in (True, None):
            end = ''
            attr = self.attributes
            if attr:
                space = ' '
            if self.on is None:
                empty = '/'
        elif self.on is False:
            end = '/'
            attr = ''
        return '<%s%s%s%s%s>' % (end, self.name, space, attr, empty)

    def __eq__(self, other):
        return self.name == other.name and self.attributes == other.attributes

    def __repr__(self):
        return "<`%s' at %d>" % (self.tag(), self.position)

class P(SilvaObject):
    """This implements the Markup type as specified in the schema"""

    # map Word styles to silva styles
    style_map = {
        'Bold': 'strong',
        'Italic': 'em',
        'Underline': 'underline',
        'Superscript': 'super',
        'Subscript': 'sub',
    }
    

    __handled_fields = [wdFieldIndexEntry, wdFieldHyperlink]

    def __init__(self, original, handle_styles=1):
        SilvaObject.__init__(self, original)

        self.text = u''
        self.markup = []
        self._non_indexes = [] 
        self._non_indexes_sorted = True

        self.handleWordParagraph(original, handle_styles)
            

    def handleWordParagraph(self, wp, handle_styles=1):
       
        def merge(list1, list2):
            """merge two insertions lists, preserving sorting""" 
            sort_dict = {}
            for tag in list1+list2:
                sort_dict.setdefault(tag.position, []).append(tag)
            indexes = sort_dict.keys()
            indexes.sort()
            result = []
            for index in indexes:
                for tag in sort_dict[index]:
                    result.append(tag)
            return result                    
        
        if wp is None: 
            return
        
        text = wp.Range()
        wp_start = wp.Range.Start
        wp_end = wp.Range.End
        
        if stripControlChars(text.strip()) == '':
            return
        xml = u''
        inserts = []

        if handle_styles:
            inserts = self.processStyles(wp)
            inserts = merge(inserts, self.processIndexes(wp))
            inserts = merge(inserts, self.processLinks(wp))
        inserts = self.wellform(inserts)
        self._blank_indexes(wp)
        self.setupKitchen()
        for tag in inserts:
            if tag.position < wp_start:
                tag.position = wp_start
            elif tag.position > wp_end:
                tag.position = wp_end
            tag.position -= wp_start
            tag.position = self.cookIndex(tag.position)

        next = 0
        for tag in inserts:
            xml += replaceMap(text[next:tag.position])
            xml += tag.tag()
            next = tag.position

        xml += replaceMap(text[next:])
        xml = p_SoftBreak.sub('<br/>', xml)
        xml = stripControlChars(xml)
                
        self.text = xml
          


    def processStyles(self, wp):
        """processes markup like bold and italic"""
       
        styles = []
        styles_dict = {}
        wp_start = wp.Range.Start
        wp_end = wp.Range.End
        
        for word_style, silva_style in self.style_map.items():
            wp_range = wp.Range
            find = wp_range.Find
            self.clearFind(find)
            setattr(find.Font, word_style, True)
            while find.Execute(Replace=wdReplaceNone):
                if wp_range.Start < wp_start: continue
                #fuzz needed for reasons i do not comprehend
                if wp_range.End +1 >= wp_end: break
                styles_dict.setdefault(wp_range.Start, []).append(
                    (self.style_map[word_style], True))
                styles_dict.setdefault(wp_range.End, []).insert(0, 
                    (self.style_map[word_style], False))
        index_list = styles_dict.keys()
        index_list.sort()
        for index in index_list:
            styles += map(lambda (style, onoff): Tag(style, '', onoff, index),
                 styles_dict[index])
        return styles

    def processIndexes(self, wp):
            
        if len(wp.Range.Fields) == 0:
            return []
    
        wp_start = wp.Range.Start
        result = []
        for field in wp.Range.Fields:
            if not field.type == wdFieldIndexEntry: continue 
            
            start = field.Code.Start
            end = field.Code.End
            
            self.nonIndex(start, end)
            
            m = p_INDEX.search(field.Code())
            if m is None: continue
            name = m.group(1)
            
            tag = Tag('index', "name='%s'" % (name, ), None, end+1)
            result.append(tag)
   
        return result


    def processLinks(self, wp):

        if len(wp.Range.Hyperlinks) == 0:
            return []
        
        wp_start = wp.Range.Start
        result = []
        
        for link in wp.Range.Hyperlinks:
            
            start = link.Range.start
            end = link.Range.end
            
            url = replaceMap(link.address)
            text = link.TextToDisplay

            text_index = end - len(text) - 2
            self.nonIndex(start, text_index-1)
            self.nonIndex(end-2, end)
            
            stag = '<link url="%s">' % (url, )
            etag = '</link>'
           
            start_tag = Tag('link', 'url="%s"' % (url, ), True, text_index)
            end_tag = copy(start_tag)
            end_tag.on = False
            end_tag.position = end-2
            result.append(start_tag)
            result.append(end_tag)
        return result

    def nonIndex(self, start, end):

        self._non_indexes += range(start, end+1)
        self._non_indexes_sorted = False
   
    def setupKitchen(self):
        """sets up the kitchen for cooking indexes"""
        
        wp_start = self.original.Range.Start
        wp_end = self.original.Range.End
        if not self._non_indexes_sorted:
            self._non_indexes = dict(map(lambda x: (x, 0), 
                self._non_indexes)).keys()
            self._non_indexes.sort()
            self._non_indexes = \
                map(lambda x: x - wp_start, 
                filter(lambda x: wp_start <= x <= wp_end,
                    self._non_indexes))
            self._non_indexes_sorted = True

   
    def cookIndex(self, raw_index):
        nis = self._non_indexes
        
        while raw_index in nis:
            raw_index += 1
        corrector = 0
        for i in range(0, len(nis)):
            if raw_index <= nis[i]:
                break
            else:
                corrector = i + 1
        return raw_index - corrector


    def _blank_indexes(self, wp):
        if len(wp.Range.Fields) == 0: 
            return
        for field in wp.Range.Fields:
            if not field.Type in self.__handled_fields:
                self.nonIndex(field.Code.Start, field.Code.End)


    def clearFind(self, find):
        """clear the given find instance"""
        find.ClearFormatting()
        find.ClearAllFuzzyOptions()
        find.Format = 1
        find.Text = u''
        find.MatchCase = 0
        find.MatchWholeWord = 0
        find.MatchWildcards = 0 
        find.MatchSoundsLike = 0
        find.MatchAllWordForms = 0
        find.Wrap = wdWrapNever 
        find.Forward = True

    def wellform(tags):
        """wellform the styles
        
            It basicaly handles cases which would lead to:

                <strong>Foo<em>Bar</strong>baz</em>

            By replacing it by:

                <strong>Foo<em>Bar</em></strong><em>baz</em>
        
            It's not done with xml tags but an internal structure though.

            returns a list of Tag instances
        """
        
        result = []
        stack = []
        for tag in tags:
            if tag.on is True:
                if not tag in stack:
                    result.append(tag)
                    stack.append(copy(tag))
            elif tag.on is False:
                if tag in stack:
                    deactivated_tags = []
                    while 1:
                        deactivate_tag = stack.pop()
                        deactivate_tag.on = False
                        deactivate_tag.position = tag.position
                        result.append(deactivate_tag)
                        if deactivate_tag == tag:
                            break
                        deactivated_tags.append(copy(deactivate_tag))
                    deactivated_tags.reverse()
                    for reactivate_tag in deactivated_tags:
                        reactivate_tag.on = True
                        reactivate_tag.position = tag.position
                        stack.append(copy(reactivate_tag))
                        result.append(reactivate_tag)
            elif tag.on is None:
                result.append(tag)
            else:
                raise ValueError, 'tag.on is invalid (%r)' % (tag.on,)
        assert stack == [], "Stack is not empty: %r (%r)" % (stack, result)
        return result
    wellform = staticmethod(wellform)

    def getXML(self):
        if self.text:
            return u'<p>%s</p>' % self.text
        return u''


class Pre(SilvaObject):

    def getXML(self):

        elements = \
            map(stripControlChars, 
            map(lambda x: replaceMap(x),
            self._elements))
        text = '\n'.join(elements)    
            
        xml = u'<pre>%s</pre>' % (text, )
        return xml

    
